<?php

use Drupal\Core\Entity\EntityLastInstalledSchemaRepositoryInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\StringTranslation\TranslatableMarkup;

/**
 * Make course enrollments revisionable.
 */
function course_post_update_make_course_enrollment_revisionable(&$sandbox) {

  $definition_update_manager = Drupal::entityDefinitionUpdateManager();
  /** @var EntityLastInstalledSchemaRepositoryInterface $last_installed_schema_repository */
  $last_installed_schema_repository = Drupal::service('entity.last_installed_schema.repository');

  $entity_type = $definition_update_manager->getEntityType('course_enrollment');
  $field_storage_definitions = $last_installed_schema_repository->getLastInstalledFieldStorageDefinitions('course_enrollment');

  // Update the entity type definition.
  $entity_keys = $entity_type->getKeys();
  $entity_keys['revision'] = 'revision_id';
  $entity_type->set('entity_keys', $entity_keys);
  $entity_type->set('revision_table', 'course_enrollment_revision');
  $revision_metadata_keys = [
    'revision_default' => 'revision_default',
    'revision_user' => 'revision_user',
    'revision_created' => 'revision_created',
    'revision_log_message' => 'revision_log_message',
  ];
  $entity_type->set('revision_metadata_keys', $revision_metadata_keys);

  // Update the field storage definitions and add the new ones required by a
  // revisionable entity type.
  $field_storage_definitions['enrollmenttype']->setRevisionable(TRUE);
  $field_storage_definitions['status']->setRevisionable(TRUE);
  $field_storage_definitions['timestamp']->setRevisionable(TRUE);
  $field_storage_definitions['enroll_end']->setRevisionable(TRUE);
  $field_storage_definitions['complete']->setRevisionable(TRUE);
  $field_storage_definitions['grade_result']->setRevisionable(TRUE);
  $field_storage_definitions['section']->setRevisionable(TRUE);
  $field_storage_definitions['section_name']->setRevisionable(TRUE);
  $field_storage_definitions['coid']->setRevisionable(TRUE);
  $field_storage_definitions['date_completed']->setRevisionable(TRUE);

  // Revision metadata
  $field_storage_definitions['revision_id'] = BaseFieldDefinition::create('integer')
    ->setName('revision_id')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision ID'))
    ->setReadOnly(TRUE)
    ->setSetting('unsigned', TRUE);

  $field_storage_definitions['revision_default'] = BaseFieldDefinition::create('boolean')
    ->setName('revision_default')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(t('Default revision'))
    ->setDescription(t('A flag indicating whether this was a default revision when it was saved.'))
    ->setRevisionable(TRUE)
    // We cannot tell whether existing revisions were default or not when
    // they were created, but since we did not support creating non-default
    // revisions in any core stable UI so far, we default to TRUE.
    ->setInitialValue(TRUE);

  $field_storage_definitions['revision_created'] = BaseFieldDefinition::create('created')
    ->setName('revision_created')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision create time'))
    ->setDescription(new TranslatableMarkup('The time that the current revision was created.'))
    ->setRevisionable(TRUE);

  $field_storage_definitions['revision_user'] = BaseFieldDefinition::create('entity_reference')
    ->setName('revision_user')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision user'))
    ->setDescription(new TranslatableMarkup('The user ID of the author of the current revision.'))
    ->setSetting('target_type', 'user')
    ->setRevisionable(TRUE);

  $field_storage_definitions['revision_log_message'] = BaseFieldDefinition::create('string_long')
    ->setName('revision_log_message')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision log message'))
    ->setDescription(new TranslatableMarkup('Briefly describe the changes you have made.'))
    ->setRevisionable(TRUE)
    ->setDefaultValue('');

  $definition_update_manager->updateFieldableEntityType($entity_type, $field_storage_definitions, $sandbox);
}

/**
 * Make courses revisionable.
 */
function course_post_update_make_courses_revisionable(&$sandbox) {
  $definition_update_manager = Drupal::entityDefinitionUpdateManager();

  // Remove these old revision fields that got accidentally added.
  if ($definition = $definition_update_manager->getFieldStorageDefinition('revision_timestamp', 'course')) {
    $definition_update_manager->uninstallFieldStorageDefinition($definition);
  }
  if ($definition = $definition_update_manager->getFieldStorageDefinition('revision_uid', 'course')) {
    $definition_update_manager->uninstallFieldStorageDefinition($definition);
  }
  if ($definition = $definition_update_manager->getFieldStorageDefinition('revision_log', 'course')) {
    $definition_update_manager->uninstallFieldStorageDefinition($definition);
  }

  /** @var EntityLastInstalledSchemaRepositoryInterface $last_installed_schema_repository */
  $last_installed_schema_repository = Drupal::service('entity.last_installed_schema.repository');

  $entity_type = $definition_update_manager->getEntityType('course');
  $field_storage_definitions = $last_installed_schema_repository->getLastInstalledFieldStorageDefinitions('course');

  // Update the entity type definition.
  $entity_keys = $entity_type->getKeys();
  $entity_keys['revision'] = 'revision_id';
  $entity_type->set('entity_keys', $entity_keys);
  $entity_type->set('revision_table', 'course_revision');
  $revision_metadata_keys = [
    'revision_default' => 'revision_default',
    'revision_user' => 'revision_user',
    'revision_created' => 'revision_created',
    'revision_log_message' => 'revision_log_message',
  ];
  $entity_type->set('revision_metadata_keys', $revision_metadata_keys);

  // UID field is new.
  $field_storage_definitions['uid'] = BaseFieldDefinition::create('entity_reference')
    ->setName('uid')
    ->setSetting('target_type', 'user')
    ->setTargetEntityTypeId('course')
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Authored by'))
    ->setDescription(new TranslatableMarkup('The username of the content author.'))
    ->setRevisionable(TRUE)
    ->setDefaultValue('');

  // Status field is new.
  $field_storage_definitions['status'] = BaseFieldDefinition::create('boolean')
    ->setName('status')
    ->setTargetEntityTypeId('course')
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Published'))
    ->setRevisionable(TRUE);

  // Update the field storage definitions and add the new ones required by a
  // revisionable entity type.
  $field_storage_definitions['course_date']->setRevisionable(TRUE);
  $field_storage_definitions['duration']->setRevisionable(TRUE);
  $field_storage_definitions['credits']->setRevisionable(TRUE);
  $field_storage_definitions['enrollment_type']->setRevisionable(TRUE);
  $field_storage_definitions['outline']->setRevisionable(TRUE);
  $field_storage_definitions['title']->setRevisionable(TRUE);

  // Revision metadata
  $field_storage_definitions['revision_id'] = BaseFieldDefinition::create('integer')
    ->setName('revision_id')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision ID'))
    ->setReadOnly(TRUE)
    ->setSetting('unsigned', TRUE);

  $field_storage_definitions['revision_default'] = BaseFieldDefinition::create('boolean')
    ->setName('revision_default')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(t('Default revision'))
    ->setDescription(t('A flag indicating whether this was a default revision when it was saved.'))
    ->setRevisionable(TRUE)
    // We cannot tell whether existing revisions were default or not when
    // they were created, but since we did not support creating non-default
    // revisions in any core stable UI so far, we default to TRUE.
    ->setInitialValue(TRUE);

  $field_storage_definitions['revision_created'] = BaseFieldDefinition::create('created')
    ->setName('revision_created')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision create time'))
    ->setDescription(new TranslatableMarkup('The time that the current revision was created.'))
    ->setRevisionable(TRUE);

  $field_storage_definitions['revision_user'] = BaseFieldDefinition::create('entity_reference')
    ->setName('revision_user')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision user'))
    ->setDescription(new TranslatableMarkup('The user ID of the author of the current revision.'))
    ->setSetting('target_type', 'user')
    ->setRevisionable(TRUE);

  $field_storage_definitions['revision_log_message'] = BaseFieldDefinition::create('string_long')
    ->setName('revision_log_message')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision log message'))
    ->setDescription(new TranslatableMarkup('Briefly describe the changes you have made.'))
    ->setRevisionable(TRUE)
    ->setDefaultValue('');

  $definition_update_manager->updateFieldableEntityType($entity_type, $field_storage_definitions, $sandbox);
}

/**
 * Add UUID field to course.
 */
function course_post_update_add_uuid_to_course(&$sandbox) {
  $definition_update_manager = Drupal::entityDefinitionUpdateManager();
  /** @var EntityLastInstalledSchemaRepositoryInterface $last_installed_schema_repository */
  $last_installed_schema_repository = Drupal::service('entity.last_installed_schema.repository');
  $entity_type = $definition_update_manager->getEntityType('course');
  $field_storage_definitions = $last_installed_schema_repository->getLastInstalledFieldStorageDefinitions('course');

  $field_storage_definitions['uuid'] = BaseFieldDefinition::create('uuid')
    ->setName('uuid')
    ->setTargetEntityTypeId('course');

  $definition_update_manager->updateFieldableEntityType($entity_type, $field_storage_definitions, $sandbox);
}

/**
 * Add UUID field to course_enrollment.
 */
function course_post_update_add_uuid_to_course_enrollment(&$sandbox) {
  $definition_update_manager = Drupal::entityDefinitionUpdateManager();
  /** @var EntityLastInstalledSchemaRepositoryInterface $last_installed_schema_repository */
  $last_installed_schema_repository = Drupal::service('entity.last_installed_schema.repository');
  $entity_type = $definition_update_manager->getEntityType('course_enrollment');
  $field_storage_definitions = $last_installed_schema_repository->getLastInstalledFieldStorageDefinitions('course_enrollment');

  $field_storage_definitions['uuid'] = BaseFieldDefinition::create('uuid')
    ->setName('uuid')
    ->setTargetEntityTypeId('course_enrollment');

  $definition_update_manager->updateFieldableEntityType($entity_type, $field_storage_definitions, $sandbox);
}

function course_post_update_add_bundle_field(&$sandbox) {
  $definition_update_manager = Drupal::entityDefinitionUpdateManager();
  /** @var EntityLastInstalledSchemaRepositoryInterface $last_installed_schema_repository */
  $last_installed_schema_repository = Drupal::service('entity.last_installed_schema.repository');
  $entity_type = $definition_update_manager->getEntityType('course');
  $field_storage_definitions = $last_installed_schema_repository->getLastInstalledFieldStorageDefinitions('course');

  $field_storage_definitions['type'] = BaseFieldDefinition::create('entity_reference')
    ->setName('type')
    ->setTargetEntityTypeId('course')
    ->setLabel($entity_type->getBundleLabel())
    ->setSetting('target_type', 'course_type')
    ->setRequired(TRUE)
    ->setReadOnly(TRUE);

  $definition_update_manager->updateFieldableEntityType($entity_type, $field_storage_definitions, $sandbox);
}

/**
 * Make course objects revisionable.
 */
function course_post_update_make_course_objects_revisionable2(&$sandbox) {

  $definition_update_manager = Drupal::entityDefinitionUpdateManager();
  /** @var EntityLastInstalledSchemaRepositoryInterface $last_installed_schema_repository */
  $last_installed_schema_repository = Drupal::service('entity.last_installed_schema.repository');

  $entity_type = $definition_update_manager->getEntityType('course_object');
  $field_storage_definitions = $last_installed_schema_repository->getLastInstalledFieldStorageDefinitions('course_object');

  // Update the entity type definition.
  $entity_keys = $entity_type->getKeys();
  $entity_keys['revision'] = 'revision_id';
  $entity_type->set('entity_keys', $entity_keys);
  $entity_type->set('base_table', 'course_object');
  $entity_type->set('data_table', NULL);
  $entity_type->set('revision_table', 'course_object_revision');
  $entity_type->set('revision_data_table', NULL);
  $revision_metadata_keys = [
    'revision_default' => 'revision_default',
    'revision_user' => 'revision_user',
    'revision_created' => 'revision_created',
    'revision_log_message' => 'revision_log_message',
  ];
  $entity_type->set('revision_metadata_keys', $revision_metadata_keys);

  // Update the field storage definitions and add the new ones required by a
  // revisionable entity type.
  $field_storage_definitions['title']->setRevisionable(TRUE);
  $field_storage_definitions['enabled']->setRevisionable(TRUE);
  $field_storage_definitions['hidden']->setRevisionable(TRUE);
  $field_storage_definitions['required']->setRevisionable(TRUE);
  $field_storage_definitions['weight']->setRevisionable(TRUE);
  $field_storage_definitions['duration']->setRevisionable(TRUE);
  $field_storage_definitions['instance']->setRevisionable(TRUE);

  // Revision metadata
  $field_storage_definitions['revision_id'] = BaseFieldDefinition::create('integer')
    ->setName('revision_id')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision ID'))
    ->setReadOnly(TRUE)
    ->setSetting('unsigned', TRUE);

  $field_storage_definitions['revision_default'] = BaseFieldDefinition::create('boolean')
    ->setName('revision_default')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(t('Default revision'))
    ->setDescription(t('A flag indicating whether this was a default revision when it was saved.'))
    ->setRevisionable(TRUE)
    // We cannot tell whether existing revisions were default or not when
    // they were created, but since we did not support creating non-default
    // revisions in any core stable UI so far, we default to TRUE.
    ->setInitialValue(TRUE);

  $field_storage_definitions['revision_created'] = BaseFieldDefinition::create('created')
    ->setName('revision_created')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision create time'))
    ->setDescription(new TranslatableMarkup('The time that the current revision was created.'))
    ->setRevisionable(TRUE);

  $field_storage_definitions['revision_user'] = BaseFieldDefinition::create('entity_reference')
    ->setName('revision_user')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision user'))
    ->setDescription(new TranslatableMarkup('The user ID of the author of the current revision.'))
    ->setSetting('target_type', 'user')
    ->setRevisionable(TRUE);

  $field_storage_definitions['revision_log_message'] = BaseFieldDefinition::create('string_long')
    ->setName('revision_log_message')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision log message'))
    ->setDescription(new TranslatableMarkup('Briefly describe the changes you have made.'))
    ->setRevisionable(TRUE)
    ->setDefaultValue('');

  $definition_update_manager->updateFieldableEntityType($entity_type, $field_storage_definitions, $sandbox);
}

/**
 * Make course object fulfillments revisionable.
 */
function course_post_update_make_course_object_fulfillments_revisionable2(&$sandbox) {

  $definition_update_manager = Drupal::entityDefinitionUpdateManager();
  /** @var EntityLastInstalledSchemaRepositoryInterface $last_installed_schema_repository */
  $last_installed_schema_repository = Drupal::service('entity.last_installed_schema.repository');

  $entity_type = $definition_update_manager->getEntityType('course_object_fulfillment');
  $field_storage_definitions = $last_installed_schema_repository->getLastInstalledFieldStorageDefinitions('course_object_fulfillment');

  // Update the entity type definition.
  $entity_keys = $entity_type->getKeys();
  $entity_keys['revision'] = 'revision_id';
  $entity_type->set('entity_keys', $entity_keys);
  $entity_type->set('base_table', 'course_object_fulfillment');
  $entity_type->set('data_table', NULL);
  $entity_type->set('revision_table', 'course_object_fulfillment_revision');
  $entity_type->set('revision_data_table', NULL);
  $revision_metadata_keys = [
    'revision_default' => 'revision_default',
    'revision_user' => 'revision_user',
    'revision_created' => 'revision_created',
    'revision_log_message' => 'revision_log_message',
  ];
  $entity_type->set('revision_metadata_keys', $revision_metadata_keys);

  // Update the field storage definitions and add the new ones required by a
  // revisionable entity type.
  $field_storage_definitions['complete']->setRevisionable(TRUE);
  $field_storage_definitions['date_completed']->setRevisionable(TRUE);
  $field_storage_definitions['grade_result']->setRevisionable(TRUE);
  $field_storage_definitions['instance']->setRevisionable(TRUE);
  $field_storage_definitions['date_started']->setRevisionable(TRUE);

  // Revision metadata
  $field_storage_definitions['revision_id'] = BaseFieldDefinition::create('integer')
    ->setName('revision_id')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision ID'))
    ->setReadOnly(TRUE)
    ->setSetting('unsigned', TRUE);

  $field_storage_definitions['revision_default'] = BaseFieldDefinition::create('boolean')
    ->setName('revision_default')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(t('Default revision'))
    ->setDescription(t('A flag indicating whether this was a default revision when it was saved.'))
    ->setRevisionable(TRUE)
    // We cannot tell whether existing revisions were default or not when
    // they were created, but since we did not support creating non-default
    // revisions in any core stable UI so far, we default to TRUE.
    ->setInitialValue(TRUE);

  $field_storage_definitions['revision_created'] = BaseFieldDefinition::create('created')
    ->setName('revision_created')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision create time'))
    ->setDescription(new TranslatableMarkup('The time that the current revision was created.'))
    ->setRevisionable(TRUE);

  $field_storage_definitions['revision_user'] = BaseFieldDefinition::create('entity_reference')
    ->setName('revision_user')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision user'))
    ->setDescription(new TranslatableMarkup('The user ID of the author of the current revision.'))
    ->setSetting('target_type', 'user')
    ->setRevisionable(TRUE);

  $field_storage_definitions['revision_log_message'] = BaseFieldDefinition::create('string_long')
    ->setName('revision_log_message')
    ->setTargetEntityTypeId($entity_type->id())
    ->setTargetBundle(NULL)
    ->setLabel(new TranslatableMarkup('Revision log message'))
    ->setDescription(new TranslatableMarkup('Briefly describe the changes you have made.'))
    ->setRevisionable(TRUE)
    ->setDefaultValue('');

  $definition_update_manager->updateFieldableEntityType($entity_type, $field_storage_definitions, $sandbox);
}

/**
 * Add UUID field to course object.
 */
function course_post_update_add_uuid_to_course_object(&$sandbox) {
  $definition_update_manager = Drupal::entityDefinitionUpdateManager();
  /** @var EntityLastInstalledSchemaRepositoryInterface $last_installed_schema_repository */
  $last_installed_schema_repository = Drupal::service('entity.last_installed_schema.repository');
  $entity_type = $definition_update_manager->getEntityType('course_object');
  $field_storage_definitions = $last_installed_schema_repository->getLastInstalledFieldStorageDefinitions('course_object');

  $field_storage_definitions['uuid'] = BaseFieldDefinition::create('uuid')
    ->setName('uuid')
    ->setTargetEntityTypeId('course_object');

  $definition_update_manager->updateFieldableEntityType($entity_type, $field_storage_definitions, $sandbox);
}

/**
 * Add UUID field to course object fulfillments.
 */
function course_post_update_add_uuid_to_course_object_fulfillment(&$sandbox) {
  $definition_update_manager = Drupal::entityDefinitionUpdateManager();
  /** @var EntityLastInstalledSchemaRepositoryInterface $last_installed_schema_repository */
  $last_installed_schema_repository = Drupal::service('entity.last_installed_schema.repository');
  $entity_type = $definition_update_manager->getEntityType('course_object_fulfillment');
  $field_storage_definitions = $last_installed_schema_repository->getLastInstalledFieldStorageDefinitions('course_object_fulfillment');

  $field_storage_definitions['uuid'] = BaseFieldDefinition::create('uuid')
    ->setName('uuid')
    ->setTargetEntityTypeId('course_object_fulfillment');

  $definition_update_manager->updateFieldableEntityType($entity_type, $field_storage_definitions, $sandbox);
}
